@@ -2,8 +2,12 @@
 
 #include <glade/glade.h>
 
+#include "gva-game-db.h"
+#include "gva-favorites.h"
 #include "gva-game-store.h"
+#include "gva-main.h"
 #include "gva-play-back.h"
+#include "gva-preferences.h"
 #include "gva-tree-view.h"
 #include "gva-util.h"
 #include "gva-xmame.h"
@@ -33,7 +37,7 @@ action_about_cb (GtkAction *action)
         if (error != NULL)
         {
                 g_warning ("%s", error->message);
-                g_error_free (error);
+                g_clear_error (&error);
         }
 
         gtk_show_about_dialog (
@@ -51,19 +55,66 @@ action_about_cb (GtkAction *action)
                 g_object_unref (logo);
 }
 
+static void
+action_auto_save_cb (GtkToggleAction *action)
+{
+        gboolean active;
+
+        active = gtk_toggle_action_get_active (action);
+        gva_preferences_set_auto_save (active);
+}
+
 static void
 action_contents_cb (GtkAction *action)
 {
 }
 
+static void
+action_full_screen_cb (GtkToggleAction *action)
+{
+        gboolean active;
+
+        active = gtk_toggle_action_get_active (action);
+        gva_preferences_set_full_screen (active);
+}
+
+static void
+action_insert_favorite_cb (GtkAction *action)
+{
+        GtkTreeModel *model;
+        GtkTreePath *path;
+        GtkTreeIter iter;
+        const gchar *romname;
+        gboolean valid;
+
+        romname = gva_tree_view_get_selected_game ();
+        g_assert (romname != NULL);
+
+        model = gva_game_db_get_model ();
+        path = gva_game_db_lookup (romname);
+        valid = gtk_tree_model_get_iter (model, &iter, path);
+        gtk_tree_path_free (path);
+        g_assert (valid);
+
+        gtk_list_store_set (
+                GTK_LIST_STORE (model), &iter,
+                GVA_GAME_STORE_COLUMN_FAVORITE, TRUE, -1);
+
+        gva_favorites_insert (romname);
+
+        gtk_action_set_visible (GVA_ACTION_INSERT_FAVORITE, FALSE);
+        gtk_action_set_visible (GVA_ACTION_REMOVE_FAVORITE, TRUE);
+}
+
 static void
 action_play_back_cb (GtkAction *action)
 {
         GtkTreeModel *model;
         GtkTreeView *view;
         GtkTreeIter iter;
-        gchar *inpname;
         gchar *inpfile;
+        gchar *inpname;
+        gchar *romname;
         GList *list;
         gboolean iter_set;
         GError *error = NULL;
@@ -80,20 +131,31 @@ action_play_back_cb (GtkAction *action)
         iter_set = gtk_tree_model_get_iter (model, &iter, list->data);
         g_assert (iter_set);
         gtk_tree_model_get (
-                model, &iter, GVA_GAME_STORE_COLUMN_INPFILE, &inpfile, -1);
+                model, &iter, GVA_GAME_STORE_COLUMN_INPFILE, &inpfile,
+                GVA_GAME_STORE_COLUMN_ROMNAME, &romname, -1);
         inpname = g_strdelimit (g_path_get_basename (inpfile), ".", '\0');
         g_free (inpfile);
 
-        if (!gva_xmame_playback_game (inpname, &error))
+        if (!gva_xmame_playback_game (romname, inpname, &error))
         {
+                g_assert (error != NULL);
                 g_warning ("%s", error->message);
-                g_error_free (error);
+                g_clear_error (&error);
         }
 
+        g_free (inpname);
+        g_free (romname);
+
         g_list_foreach (list, (GFunc) gtk_tree_path_free, NULL);
         g_list_free (list);
 }
 
+static void
+action_preferences_cb (GtkAction *action)
+{
+        gtk_widget_show (GVA_WIDGET_PREFERENCES_WINDOW);
+}
+
 static void
 action_properties_cb (GtkAction *action)
 {
@@ -116,14 +178,43 @@ action_record_cb (GtkAction *action)
         g_assert (romname != NULL);
 
         inpname = gva_choose_inpname (romname);
-        gva_xmame_record_game (romname, inpname, &error);
-        g_free (inpname);
 
-        if (error != NULL)
+        if (!gva_xmame_record_game (romname, inpname, &error))
         {
+                g_assert (error != NULL);
                 g_warning ("%s", error->message);
                 g_clear_error (&error);
         }
+
+        g_free (inpname);
+}
+
+static void
+action_remove_favorite_cb (GtkAction *action)
+{
+        GtkTreeModel *model;
+        GtkTreePath *path;
+        GtkTreeIter iter;
+        const gchar *romname;
+        gboolean valid;
+
+        romname = gva_tree_view_get_selected_game ();
+        g_assert (romname != NULL);
+
+        model = gva_game_db_get_model ();
+        path = gva_game_db_lookup (romname);
+        valid = gtk_tree_model_get_iter (model, &iter, path);
+        gtk_tree_path_free (path);
+        g_assert (valid);
+
+        gtk_list_store_set (
+                GTK_LIST_STORE (model), &iter,
+                GVA_GAME_STORE_COLUMN_FAVORITE, FALSE, -1);
+
+        gva_favorites_remove (romname);
+
+        gtk_action_set_visible (GVA_ACTION_INSERT_FAVORITE, TRUE);
+        gtk_action_set_visible (GVA_ACTION_REMOVE_FAVORITE, FALSE);
 }
 
 static void
@@ -141,10 +232,19 @@ action_start_cb (GtkAction *action)
         romname = gva_tree_view_get_selected_game ();
         g_assert (romname != NULL);
 
-        gva_xmame_run_game (romname, &error);
-
-        if (error != NULL)
+        if (!gva_preferences_get_auto_save ())
         {
+                if (!gva_xmame_clear_state (romname, &error))
+                {
+                        g_assert (error != NULL);
+                        g_warning ("%s", error->message);
+                        g_clear_error (&error);
+                }
+        }
+
+        if (!gva_xmame_run_game (romname, &error))
+        {
+                g_assert (error != NULL);
                 g_warning ("%s", error->message);
                 g_clear_error (&error);
         }
@@ -158,18 +258,13 @@ action_view_changed_cb (GtkRadioAction *action, GtkRadioAction *current)
                 gva_tree_view_get_selected_view ());
 }
 
-static void
-action_unmark_favorite_cb (GtkAction *action)
-{
-}
-
 static GtkActionEntry entries[] =
 {
         { "about",
           GTK_STOCK_ABOUT,
           N_("_About"),
           NULL,
-          NULL,
+          N_("Show information about the application"),
           G_CALLBACK (action_about_cb) },
 
         { "contents",
@@ -179,6 +274,13 @@ static GtkActionEntry entries[] =
           NULL,
           G_CALLBACK (action_contents_cb) },
 
+        { "insert-favorite",
+          GTK_STOCK_ADD,
+          N_("Add to _Favorites"),
+          "<Control>plus",
+          N_("Add the selected game to my list of favorites"),
+          G_CALLBACK (action_insert_favorite_cb) },
+
         { "play-back",
           GTK_STOCK_MEDIA_PLAY,
           N_("Play _Back..."),
@@ -186,18 +288,25 @@ static GtkActionEntry entries[] =
           N_("Play back the selected game recording"),
           G_CALLBACK (action_play_back_cb) },
 
+        { "preferences",
+          GTK_STOCK_PREFERENCES,
+          N_("_Preferences"),
+          NULL,
+          N_("Configure the application"),
+          G_CALLBACK (action_preferences_cb) },
+
         { "properties",
           GTK_STOCK_PROPERTIES,
           N_("_Properties"),
           NULL,
-          N_("Show properties about the selected game"),
+          N_("Show information about the selected game"),
           G_CALLBACK (action_properties_cb) },
 
         { "quit",
           GTK_STOCK_QUIT,
           N_("_Quit"),
           NULL,
-          NULL,
+          N_("Quit the application"),
           G_CALLBACK (action_quit_cb) },
 
         { "record",
@@ -207,6 +316,13 @@ static GtkActionEntry entries[] =
           N_("Start the selected game and record keypresses to a file"),
           G_CALLBACK (action_record_cb) },
 
+        { "remove-favorite",
+          GTK_STOCK_REMOVE,
+          N_("Remove from _Favorites"),
+          "<Control>minus",
+          N_("Remove the selected game from my list of favorites"),
+          G_CALLBACK (action_remove_favorite_cb) },
+
         { "show-play-back",
           GTK_STOCK_MEDIA_PLAY,
           N_("Play _Back..."),
@@ -221,6 +337,13 @@ static GtkActionEntry entries[] =
           N_("Start the selected game"),
           G_CALLBACK (action_start_cb) },
 
+        { "edit-menu",
+          NULL,
+          N_("_Edit"),
+          NULL,
+          NULL,
+          NULL },
+
         { "game-menu",
           NULL,
           N_("_Game"),
@@ -243,6 +366,25 @@ static GtkActionEntry entries[] =
           NULL }
 };
 
+static GtkToggleActionEntry toggle_entries[] =
+{
+        { "auto-save",
+          NULL,
+          N_("_Restore previous state when starting a game"),
+          NULL,
+          NULL,
+          G_CALLBACK (action_auto_save_cb),
+          FALSE },  /* GConf overrides this */
+
+        { "full-screen",
+          NULL,
+          N_("Start games in _fullscreen mode"),
+          NULL,
+          NULL,
+          G_CALLBACK (action_full_screen_cb),
+          FALSE }   /* GConf overrides this */
+};
+
 static GtkRadioActionEntry view_radio_entries[] =
 {
         { "view-available",
@@ -272,10 +414,19 @@ gva_ui_init (void)
 {
         GtkWidget *widget;
         gchar *filename;
+        GError *error = NULL;
+
+        /* Set this immediately so we don't cause infinite recursion
+         * when we use the GVA_WIDGET_MAIN_WINDOW macro below. */
+        initialized = TRUE;
 
         action_group = gtk_action_group_new ("main");
         gtk_action_group_add_actions (
-                action_group, entries, G_N_ELEMENTS (entries), NULL);
+                action_group, entries,
+                G_N_ELEMENTS (entries), NULL);
+        gtk_action_group_add_toggle_actions (
+                action_group, toggle_entries,
+                G_N_ELEMENTS (toggle_entries), NULL);
         gtk_action_group_add_radio_actions (
                 action_group, view_radio_entries,
                 G_N_ELEMENTS (view_radio_entries),
@@ -290,15 +441,25 @@ gva_ui_init (void)
         if (filename != NULL)
         {
                 manager = gtk_ui_manager_new ();
-                gtk_ui_manager_add_ui_from_file (manager, filename, NULL);
                 gtk_ui_manager_insert_action_group (manager, action_group, 0);
+
+                gtk_window_add_accel_group (
+                        GTK_WINDOW (GVA_WIDGET_MAIN_WINDOW),
+                        gtk_ui_manager_get_accel_group (manager));
+
+                g_signal_connect (
+                        manager, "connect-proxy",
+                        G_CALLBACK (gva_main_connect_proxy_cb), NULL);
+
+                gtk_ui_manager_add_ui_from_file (manager, filename, &error);
         }
         g_free (filename);
 
+        if (error != NULL)
+                g_error ("%s", error->message);
+
         if (xml == NULL || manager == NULL)
                 g_error ("%s", _("Failed to initialize user interface"));
-
-        initialized = TRUE;
 }
 
 GtkAction *
@@ -317,7 +478,7 @@ gva_ui_get_action (const gchar *action_name)
 }
 
 GtkWidget *
-gva_ui_get_glade_widget (const gchar *widget_name)
+gva_ui_get_widget (const gchar *widget_name)
 {
         GtkWidget *widget;
 
